local sw, sh = application:getContentWidth(), application:getContentHeight()
-- a blue gradient background for the sky
local sky = Pixel.new(0xFFFFFF, 1, 320 * 3, 480 * 3)
sky:setAnchorPoint(0.5, 0.5)
sky:setColor(0x00FFFF, 1, 0x0040FF, 1, 90)
sky:setPosition(sw / 2, sh / 2)
stage:addChild(sky)

-- react physics 3D
r3d = require "reactphysics3d"
local world = r3d.World.new(0, -9.8, 0)
-- a fullscreen 3D viewport
local view = D3.View.new(sw, sh, 45, 0.1, 512) -- function D3View:init(sw, sh, fov, near, far)

-- keyboard control
local isup, isdown, istorqueleft, istorqueright, isstrifeleft, isstriferight = false, false, false, false, false, false
local isjump = false

-- a floor plane (with normals)
local gplane = D3.Mesh.new()
local gplanesize = 512
gplane:setVertexArray{	-gplanesize,1,-gplanesize,	gplanesize,1,-gplanesize,	gplanesize,1,gplanesize,	-gplanesize,1,gplanesize	}
-- Large texture coordinates, the texture will repeat
local tw, th = 3200, 3200 -- 3200, 3200
gplane:setTextureCoordinateArray{0, 0, tw, 0, tw, th, 0, th}
gplane:setGenericArray(3, Shader.DFLOAT, 3, 4, {	0,1,0,	0,1,0,	0,1,0,	0,1,0,	}) -- the normal array
gplane:setIndexArray{1,2,3,1,3,4}
gplane:setTexture(Texture.new("3d/grass.png", true, {wrap = TextureBase.REPEAT}))
gplane:updateMode(D3.Mesh.MODE_LIGHTING|D3.Mesh.MODE_SHADOW|D3.Mesh.MODE_TEXTURE) -- lighting, shadows, and texture since setTexture doesn't automatically do that
gplane.body = world:createBody()
gplane.body:setType(r3d.Body.STATIC_BODY)
gplane.body:createFixture(r3d.BoxShape.new(gplanesize, 1, gplanesize), nil, 1000) -- 1000 = ground weight

-- another ground
local ground2 = loadObj("3d", "mapx02.obj")
ground2.shape = r3d.BoxShape.new(50, 0.1, 50)
ground2.body = world:createBody()
ground2.body:setType(r3d.Body.STATIC_BODY)
ground2.body:createFixture(ground2.shape, nil, 1000)

world.bodies = {}

-- a generic CUBE Mesh with normals
local GENCUBE = D3.Cube.new()
GENCUBE:mapTexture(Texture.new("3d/box.png", true))
GENCUBE:updateMode(D3.Mesh.MODE_LIGHTING|D3.Mesh.MODE_SHADOW)
GENCUBE.shape = r3d.BoxShape.new(1, 1, 1)

function build_cube(size)
	local v = Viewport.new()
	v:setContent(GENCUBE)
	v:setPosition(math.random() * 9 - 5, math.random() * 10 + 10, math.random() * 9 - 5)
	v:setRotation(math.random() * 20 - 10)
	v:setRotationX(math.random() * 20 - 10)
	v:setRotationY(math.random() * 20 - 10)
	v.body = world:createBody(v:getMatrix())
	v.body:createFixture(GENCUBE.shape, nil, 1)
	world.bodies[v] = v.body
	return v
end

-- a ship
local ship = loadObj("3d", "ship.obj")
ship.shape = r3d.BoxShape.new(2, 1, 3)
local v2 = Viewport.new()
--v2:setRotationY(270)
v2:setContent(ship)
v2:setPosition(0, 10, -8) -- min 2 for y
v2.body = world:createBody(v2:getMatrix())
v2.body:createFixture(ship.shape, nil, 100)
--v2.body:enableGravity(false)

-- the scene
local scene = view:getScene() -- a 3D View sprite
scene:addChild(gplane)
scene:addChild(ground2)
scene:addChild(ship)
Lighting.setLight(15, 30, 0, 0.3)
Lighting.setLightTarget(0, 0, 0, 40, 120)
view:lookAt(0, 10, -20, 0, 5, 0)
stage:addChild(view)

-- Main loop
local gen = 0
local m = v2.body:getTransform()
local ax, ay, az = m:transformPoint(0, 0, 0)
local bx, by, bz = m:transformPoint(0, 0, 1)
local dx, dy, dz = bx - ax, by - ay, bz - az
local camx, camy, camz
stage:addEventListener(Event.ENTER_FRAME, function(e)
	Lighting.computeShadows(scene)
	-- the boxes
	gen = gen + 1
	if gen == 200 then
		scene:addChild(build_cube(1))	
		gen = 0
	end
	world:step(e.deltaTime)
	for s, b in pairs(world.bodies) do
		s:setMatrix(b:getTransform())
	end
	-- the ship
	m = v2.body:getTransform()
	ax, ay, az = m:transformPoint(0, 0, 0)
	bx, by, bz = m:transformPoint(0, 0, 1)
	dx, dy, dz = bx - ax, by - ay, bz - az
	if isup then v2.body:applyForce(-dx * 12 * 64, 0, -dz * 12 * 64) end
	if isdown then v2.body:applyForce(dx * 6 * 64, 0, dz * 6 * 64) end
	if istorqueleft then v2.body:applyTorque(0, -10 * 64, 0) end
	if istorqueright then v2.body:applyTorque(0, 10 * 64, 0) end
	if isstriferight then v2.body:applyForce(10 * 64, 0, 0) end
	if isstrifeleft then v2.body:applyForce(-10 * 64, 0, 0) end
	if isjump then v2.body:applyForce(0, 1024 * 64, 0) isjump = false end
	ship:setMatrix(v2.body:getTransform())
	-- the camera
	camx, camy, camz = ship:getX(), ship:getY() + 16, ship:getZ()
	-- place the camera behind the player
	camx += dx * 16 -- dx * 8
	camy += -10 -- +4
	camz += dz * 10 -- dz * 4
	-- FPS STYLE
--	view:lookAt(camx, camy, camz, ship:getX(), ship:getY() + 2, ship:getZ(),	0, 1, 0)
	view:lookAt(camx, camy, camz, ship:getX() + 0, ship:getY() + 2, ship:getZ() + 0,	0, 1, 0)
end)

-- keyboard
function onKeyDown(e)
	if e.keyCode == KeyCode.UP then isup = true end
	if e.keyCode == KeyCode.DOWN then isdown = true end
	if e.keyCode == KeyCode.LEFT then istorqueleft = true end
	if e.keyCode == KeyCode.RIGHT then istorqueright = true end
	if e.keyCode == KeyCode.J then isstrifeleft = true end
	if e.keyCode == KeyCode.L then isstriferight = true end
	if e.keyCode == KeyCode.SPACE then isjump = true end
end
function onKeyUp(e)
	if e.keyCode == KeyCode.UP then isup = false end
	if e.keyCode == KeyCode.DOWN then isdown = false end
	if e.keyCode == KeyCode.LEFT then istorqueleft = false end
	if e.keyCode == KeyCode.RIGHT then istorqueright = false end
	if e.keyCode == KeyCode.J then isstrifeleft = false end
	if e.keyCode == KeyCode.L then isstriferight = false end
	if e.keyCode == KeyCode.SPACE then isjump = false end
end
stage:addEventListener(Event.KEY_DOWN, onKeyDown)
stage:addEventListener(Event.KEY_UP, onKeyUp)
